
__author__ = 'cchatel'

import numpy as np
import queue
from sklearn import cluster
from sklearn.base import ClassifierMixin


import cProfile

import cython
cimport numpy as np


from sklearn.datasets import load_digits
iris = load_digits()
from sklearn.model_selection import train_test_split
iris_train, iris_test, iris_train_target, iris_test_target = train_test_split(iris.data, iris.target, test_size=0.33)


cpdef out_value(classes):
        if np.all(classes == classes[0]):
            return classes[0]
        else:
            return None




cdef class Node:
    cdef np.ndarray center
    cdef children
    cdef np.ndarray examples_indices	
    cdef n_features_used
    cdef np.ndarray features_indices_used
    cdef public out 
    def __init__(self, examples, classes, examples_indices, n_features_used=None):
        self.center = np.mean(examples[examples_indices], axis=0)
        self.examples_indices = examples_indices
        self.out = out_value(classes[examples_indices])
        self.children = [None, None]
        n_features = examples.shape[1]
        if n_features_used is None:
            self.n_features_used = n_features
            self.features_indices_used = np.arange(n_features)
        else:
            self.n_features_used = n_features_used
            self.features_indices_used = np.random.choice(n_features, n_features_used, replace=False)

   
    cpdef left_child(self):
        return self.children[0]


    cpdef right_child(self):
        return self.children[1]



    cpdef split(self, examples, classes):
        k_means = cluster.KMeans(n_clusters=2)
        node_examples = examples[self.examples_indices][:, self.features_indices_used]
        kmean_prediction = k_means.fit_predict(node_examples)
        left_indices = self.examples_indices[kmean_prediction == 0]
        right_indices = self.examples_indices[kmean_prediction == 1]
        while len(right_indices) == 0 or len(left_indices) == 0:  # happens when two examples are identical on the
            # chosen features
            self.features_indices_used = np.random.choice(examples.shape[1], self.n_features_used, replace=False)
            node_examples = examples[self.examples_indices][:, self.features_indices_used]
            kmean_prediction = k_means.fit_predict(node_examples)
            left_indices = self.examples_indices[kmean_prediction == 0]
            right_indices = self.examples_indices[kmean_prediction == 1]
        left_child = Node(examples, classes, left_indices, self.n_features_used)
        right_child = Node(examples, classes, right_indices, self.n_features_used)
        self.children = (left_child, right_child)


    cpdef propagate(self, examples, propagation):
        if propagation == 'all':
            distances_to_left = np.array(
                [np.linalg.norm(example - self.children[0].center) for example in examples])
            distances_to_right = np.array(
                [np.linalg.norm(example - self.children[1].center) for example in examples])
        elif propagation == 'used':
            distances_to_left = np.linalg.norm(examples[:, self.features_indices_used] - np.tile(
                self.children[0].center[self.features_indices_used], (examples.shape[0], 1)), axis=1)
            distances_to_right = np.linalg.norm(examples[:, self.features_indices_used] - np.tile(
                self.children[1].center[self.features_indices_used], (examples.shape[0], 1)), axis=1)
        propagate_left = np.arange(len(examples))[distances_to_left < distances_to_right]
        propagate_right = np.arange(len(examples))[~(distances_to_left < distances_to_right)]
        return propagate_left, propagate_right

    cdef __eq__(self, other):
        return np.array_equal(self.center, other.center) and self.out == other.out
    def __repr__(self):
        return super().__repr__()


cdef class KMeansDecisionTree :

    cdef float max_features
    cdef int mode_count
    cdef int n_leaves
    cdef int depth
    cdef root
    cdef n_features_used
    cdef str propagation



    def __init__(self, max_features: object = 1., propagation: object = 'used'):
        self.root = None
        self.max_features = max_features
        self.n_features_used = None
        self.propagation = propagation

    cdef reset(self):
        self.root = None

    cdef n_features_to_use(self, n_features):
        if isinstance(self.max_features, int):
            n_features_used = self.max_features
        elif isinstance(self.max_features, float):
            n_features_used = int(n_features * self.max_features)
            if n_features_used == 0:
                n_features_used = 1
        else:
            raise ValueError('max_features must be int or float')
        return n_features_used


    cpdef fit(self, examples, classes):
        self.reset()
        cdef int n_features
        n_examples, n_features = examples.shape
        self.n_features_used = self.n_features_to_use(n_features)
        self.root = Node(examples, classes, np.arange(len(examples)), self.n_features_used)
        node_queue = queue.Queue()
        node_queue.put(self.root)
        while not node_queue.empty():
            current_node = node_queue.get()
            current_node.split(examples, classes)
            left_child = current_node.left_child()
            right_child = current_node.right_child()
            if left_child.out is None:
                node_queue.put(left_child)
            if right_child.out is None:
                node_queue.put(right_child)

    cdef predict(self, examples):
        node_indices_queue = queue.Queue()
        root = self.root
        nb_examples = examples.shape[0]
        out = np.empty(nb_examples, dtype='int')
        root_indices = np.arange(nb_examples)
        node_indices_queue.put((root, root_indices))
        while not node_indices_queue.empty():
            node, node_indices = node_indices_queue.get()
            if node.out is None:
                left, right = node.propagate(examples[node_indices], self.propagation)
                node_indices_queue.put((node.left_child(), node_indices[left]))
                node_indices_queue.put((node.right_child(), node_indices[right]))
            else:
                out[node_indices] = node.out
        return out



    cdef classes_by_level(self, examples):
        """

        :param examples: array
            examples to classify
        :return: array of arrays
            an array of classes for each level
        """
        level_queue = queue.Queue()
        level_queue.put(([self.root], [np.arange(examples.shape[0])]))
        classes = []
        while not level_queue.empty():
            current_level, examples_indices_level = level_queue.get()
            level_classes = []
            next_level_nodes = []
            next_level_examples_indices = []
            copied_nodes = []
            copied_examples_indices = []
            for i in range(len(current_level)):
                node = current_level[i]
                examples_indices = examples_indices_level[i]
                level_classes.append(examples_indices)
                if node.left_child() is not None and node.right_child() is not None:
                    next_level_nodes.append(node.left_child())
                    next_level_nodes.append(node.right_child())
                    propagate_left, propagate_right = node.propagate(examples[examples_indices], self.propagation)
                    next_level_examples_indices.append(examples_indices[propagate_left])
                    next_level_examples_indices.append(examples_indices[propagate_right])
                else:
                    copied_nodes.append(node)
                    copied_examples_indices.append(examples_indices)
            if len(next_level_nodes) != 0:
                level_queue.put(
                    (next_level_nodes + copied_nodes, next_level_examples_indices + copied_examples_indices))
            classes.append(level_classes)
        return classes



    cdef n_leaves(self):
        nb_leaves = 0
        node_queue = queue.Queue()
        node_queue.put(self.root)
        while not node_queue.empty():
            current_node = node_queue.get()
            left_child = current_node.left_child()
            right_child = current_node.right_child()
            if left_child is None and right_child is None:
                nb_leaves += 1
            else:
                node_queue.put(left_child)
                node_queue.put(right_child)
        return nb_leaves


    cdef node_count(self):
        cdef int nb_nodes
        nb_nodes = 0
        node_queue = queue.Queue()
        node_queue.put(self.root)
        while not node_queue.empty():
            current_node = node_queue.get()
            left_child = current_node.left_child()
            right_child = current_node.right_child()
            nb_nodes += 1
            if left_child is not None and right_child is not None:
                node_queue.put(left_child)
                node_queue.put(right_child)
        return nb_nodes


    cdef depth(self):
        def recursive_depth(root):
            children = root.children
            if children == [None, None]:
                return 1
            else:
                left_child, right_child = children
                left_depth = recursive_depth(left_child)
                right_depth = recursive_depth(right_child)
            return 1 + max(left_depth, right_depth)

        return recursive_depth(self.root)
