import numpy
import queue
from sklearn import cluster
from sklearn.base import BaseEstimator, ClassifierMixin


class KMeansForest(BaseEstimator, ClassifierMixin):
    """
        A KMeans decision trees ensemble.

        :param n_estimators: int
            The number of trees to fit
        :param max_features: int, float (default=0.75)
            The number of features to consider when splitting the nodes
            If int, consider max_features at each split
            If float, max_features is a percentage and int(n_features * self.max_features) are considered at each split
        :param propagation: string
            The method of propagation used when predicting an example class.
            If 'used', propagate in the subspace used at each node
            If 'all', propagate in the entire space

        Attributes
        ----------
        estimators : array
            The array of fitted estimators

        Properties
        ----------
        node_count : int
            The total number of nodes of the estimators
    """

    def __init__(self, n_estimators=10, max_features=0.75, propagation='used'):
        self.n_estimators = n_estimators
        self.estimators = numpy.array([])
        self.max_features = max_features
        self.propagation = propagation

    def reset(self):
        self.estimators = numpy.array([])

    def fit(self, examples, classes):
        self.reset()
        for estimator_index in range(self.n_estimators):
            self.estimators = numpy.append(self.estimators, [KMeansDecisionTree(self.max_features, self.propagation)])
            self.estimators[estimator_index].fit(examples, classes)

    def predict(self, examples):
        if len(self.estimators) == 0:
            raise ValueError('Estimator not fitted')
        predictions = numpy.array([tree.predict(examples) for tree in self.estimators], dtype='int')
        counts = [numpy.bincount(predictions[:, i]) for i in range(predictions.shape[1])]
        return numpy.array([numpy.argmax(count_example) for count_example in counts])

    @property
    def node_count(self):
        return sum([tree.node_count for tree in self.estimators])


class KMeansDecisionTree(BaseEstimator, ClassifierMixin):
    """
        A KMeans decision tree.

        :param max_features: int, float (default=1.)
            The number of features to consider when splitting the nodes
            If int, consider max_features at each split
            If float, max_features is a percentage and int(n_features * self.max_features) are considered at each split
        :param propagation: string
            The method of propagation used when predicting an example class.
            If 'used', propagate in the subspace used at each node
            If 'all', propagate in the entire space


        Properties
        ----------
        node_count : int
            The total number of nodes of the estimator
        n_leaves : int
        depth : int
    """

    def __init__(self, max_features=1., propagation='used'):
        self.root = None
        self.max_features = max_features
        self.n_features_used = None
        self.propagation = propagation

    def reset(self):
        self.root = None

    def n_features_to_use(self, n_features):
        if isinstance(self.max_features, int):
            n_features_used = self.max_features
        elif isinstance(self.max_features, float):
            n_features_used = int(n_features * self.max_features)
            if n_features_used == 0:
                n_features_used = 1
        else:
            raise ValueError('max_features must be int or float')
        return n_features_used

    # @profile
    def fit(self, examples, classes):
        self.reset()
        n_examples, n_features = examples.shape
        self.n_features_used = self.n_features_to_use(n_features)
        self.root = Node(examples, classes, numpy.arange(len(examples)), self.n_features_used)
        node_queue = queue.Queue()
        node_queue.put(self.root)
        while not node_queue.empty():
            current_node = node_queue.get()
            current_node.split(examples, classes)
            left_child = current_node.left_child
            right_child = current_node.right_child
            if left_child.out is None:
                node_queue.put(left_child)
            if right_child.out is None:
                node_queue.put(right_child)

    def predict(self, examples):
        node_indices_queue = queue.Queue()
        root = self.root
        nb_examples = examples.shape[0]
        out = numpy.empty(nb_examples, dtype='int')
        root_indices = numpy.arange(nb_examples)
        node_indices_queue.put((root, root_indices))
        while not node_indices_queue.empty():
            node, node_indices = node_indices_queue.get()
            if node.out is None:
                left, right = node.propagate(examples[node_indices], self.propagation)
                node_indices_queue.put((node.left_child, node_indices[left]))
                node_indices_queue.put((node.right_child, node_indices[right]))
            else:
                out[node_indices] = node.out
        return out

    def classes_by_level(self, examples):
        """

        :param examples: array
            examples to classify
        :return: array of arrays
            an array of classes for each level
        """
        level_queue = queue.Queue()
        level_queue.put(([self.root], [numpy.arange(examples.shape[0])]))
        classes = []
        while not level_queue.empty():
            current_level, examples_indices_level = level_queue.get()
            level_classes = []
            next_level_nodes = []
            next_level_examples_indices = []
            copied_nodes = []
            copied_examples_indices = []
            for i in range(len(current_level)):
                node = current_level[i]
                examples_indices = examples_indices_level[i]
                level_classes.append(examples_indices)
                if node.left_child is not None and node.right_child is not None:
                    next_level_nodes.append(node.left_child)
                    next_level_nodes.append(node.right_child)
                    propagate_left, propagate_right = node.propagate(examples[examples_indices], self.propagation)
                    next_level_examples_indices.append(examples_indices[propagate_left])
                    next_level_examples_indices.append(examples_indices[propagate_right])
                else:
                    copied_nodes.append(node)
                    copied_examples_indices.append(examples_indices)
            if len(next_level_nodes) != 0:
                level_queue.put(
                    (next_level_nodes + copied_nodes, next_level_examples_indices + copied_examples_indices))
            classes.append(level_classes)
        return classes

    @property
    def n_leaves(self):
        nb_leaves = 0
        node_queue = queue.Queue()
        node_queue.put(self.root)
        while not node_queue.empty():
            current_node = node_queue.get()
            left_child = current_node.left_child
            right_child = current_node.right_child
            if left_child is None and right_child is None:
                nb_leaves += 1
            else:
                node_queue.put(left_child)
                node_queue.put(right_child)
        return nb_leaves

    @property
    def node_count(self):
        nb_nodes = 0
        node_queue = queue.Queue()
        node_queue.put(self.root)
        while not node_queue.empty():
            current_node = node_queue.get()
            left_child = current_node.left_child
            right_child = current_node.right_child
            nb_nodes += 1
            if left_child is not None and right_child is not None:
                node_queue.put(left_child)
                node_queue.put(right_child)
        return nb_nodes

    @property
    def depth(self):
        def recursive_depth(root):
            children = root.children
            if children == [None, None]:
                return 1
            else:
                left_child, right_child = children
                left_depth = recursive_depth(left_child)
                right_depth = recursive_depth(right_child)
            return 1 + max(left_depth, right_depth)

        return recursive_depth(self.root)


class Node:
    def __init__(self, examples, classes, examples_indices, n_features_used=None):
        self.center = numpy.mean(examples[examples_indices], axis=0)
        self.examples_indices = examples_indices
        self.out = Node.out_value(classes[examples_indices])
        self.children = [None, None]
        n_features = examples.shape[1]
        if n_features_used is None:
            self.n_features_used = n_features
            self.features_indices_used = numpy.arange(n_features)
        else:
            self.n_features_used = n_features_used
            self.features_indices_used = numpy.random.choice(n_features, n_features_used, replace=False)

    @property
    def left_child(self):
        return self.children[0]

    @property
    def right_child(self):
        return self.children[1]

    @staticmethod
    def out_value(classes):
        if numpy.all(classes == classes[0]):
            return classes[0]
        else:
            return None

    def split(self, examples, classes):
        k_means = cluster.KMeans(n_clusters=2)
        node_examples = examples[self.examples_indices][:, self.features_indices_used]
        kmean_prediction = k_means.fit_predict(node_examples)
        left_indices = self.examples_indices[kmean_prediction == 0]
        right_indices = self.examples_indices[kmean_prediction == 1]
        while len(right_indices) == 0 or len(left_indices) == 0:  # happens when two examples are identical on the
            # chosen features
            self.features_indices_used = numpy.random.choice(examples.shape[1], self.n_features_used, replace=False)
            node_examples = examples[self.examples_indices][:, self.features_indices_used]
            kmean_prediction = k_means.fit_predict(node_examples)
            left_indices = self.examples_indices[kmean_prediction == 0]
            right_indices = self.examples_indices[kmean_prediction == 1]
        left_child = Node(examples, classes, left_indices, self.n_features_used)
        right_child = Node(examples, classes, right_indices, self.n_features_used)
        self.children = (left_child, right_child)
        del self.examples_indices

    def propagate(self, examples, propagation):
        if propagation == 'all':
            distances_to_left = numpy.array(
                [numpy.linalg.norm(example - self.children[0].center) for example in examples])
            distances_to_right = numpy.array(
                [numpy.linalg.norm(example - self.children[1].center) for example in examples])
        elif propagation == 'used':
            distances_to_left = numpy.linalg.norm(examples[:, self.features_indices_used] - numpy.tile(
                self.children[0].center[self.features_indices_used], (examples.shape[0], 1)), axis=1)
            distances_to_right = numpy.linalg.norm(examples[:, self.features_indices_used] - numpy.tile(
                self.children[1].center[self.features_indices_used], (examples.shape[0], 1)), axis=1)
        propagate_left = numpy.arange(len(examples))[distances_to_left < distances_to_right]
        propagate_right = numpy.arange(len(examples))[~(distances_to_left < distances_to_right)]
        return propagate_left, propagate_right

    def __eq__(self, other):
        return numpy.array_equal(self.center, other.center) and self.out == other.out
