import numpy
import unittest
from kmeansdt import Node, KMeansDecisionTree, KMeansForest


class TestException(unittest.TestCase):
    # Classe Node que split et propagate teste

    def test_split(self):
        examples = numpy.array([[0, 0], [0, 1], [3, 0], [3, 1]])
        classes = numpy.array([0, 0, 1, 1])
        initial_node = Node(examples, classes, numpy.arange(examples.shape[0]))
        first_child = Node(examples, classes, numpy.array([0, 1]))
        second_child = Node(examples, classes, numpy.array([2, 3]))
        initial_node.split(examples, classes)
        assert (initial_node.left_child == first_child and initial_node.right_child == second_child) or (
            initial_node.left_child == second_child and initial_node.right_child == first_child)

    def test_propagate(self):
        examples = numpy.array([[0, 0], [0, 1], [3, 0], [3, 1]])
        classes = numpy.array([0, 0, 1, 1])
        initial_node = Node(examples, classes, numpy.arange(examples.shape[0]))
        first_child = Node(examples, classes, numpy.array([0, 1]))
        second_child = Node(examples, classes, numpy.array([2, 3]))
        initial_node.children[0] = first_child
        initial_node.children[1] = second_child
        assert numpy.array_equal(
            initial_node.propagate(numpy.array([[0, 0], [0, 1], [3, 0], [3, 1], [3, 3], [0.1, 0.7]]),
                                   propagation='all')[0],
            [0, 1, 5])
        assert numpy.array_equal(
            initial_node.propagate(numpy.array([[0, 0], [0, 1], [3, 0], [3, 1], [3, 3], [0.1, 0.7]]),
                                   propagation='all')[1],
            [2, 3, 4])
        assert numpy.array_equal(
            initial_node.propagate(numpy.array([[0, 0], [0, 1], [3, 0], [3, 1], [3, 3], [0.1, 0.7]]),
                                   propagation='used')[0],
            [0, 1, 5])
        assert numpy.array_equal(
            initial_node.propagate(numpy.array([[0, 0], [0, 1], [3, 0], [3, 1], [3, 3], [0.1, 0.7]]),
                                   propagation='used')[1],
            [2, 3, 4])

    # Tests tree
    def test_init(self):
        tree = KMeansDecisionTree()
        assert tree.root is None
        assert tree.max_features == 1.
        assert tree.n_features_used is None
        assert tree.propagation == 'used'

    def test_reset(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [2, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 1, 1])
        tree = KMeansDecisionTree()
        tree.fit(example, example_classes)
        tree.reset()
        assert tree.root is None

    def test_maxfeature(self):
        kmdt = KMeansDecisionTree(1)
        assert kmdt.n_features_to_use(0) == 1
        kmdt = KMeansDecisionTree(0.75)
        assert kmdt.n_features_to_use(0) == 1
        assert kmdt.n_features_to_use(1) == 1

    def test_raise_maxfeature(self):
        kmdt = KMeansDecisionTree("feature")
        with self.assertRaises(ValueError):
            kmdt.n_features_to_use(1)

    def test_tree_fit(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [2, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 1, 1])
        tree = KMeansDecisionTree()
        tree.fit(example, example_classes)
        assert tree.root.out is None
        assert tree.root.left_child.out == 0 and tree.root.left_child.out == 1
        assert tree.root.right_child.out == 1 or tree.root.right_child.out == 0
        assert tree.root.left_child.out != tree.root.right_child.out
        assert tree.root.left_child.children == [None, None]
        assert tree.root.right_child.children == [None, None]

    def test_fit_depth_two(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        root = Node(example, example_classes, [0, 1, 2, 3, 4, 5])
        left_child = Node(example, example_classes, [0, 1, 2])
        right_child = Node(example, example_classes, [3, 4, 5])
        right_left = Node(example, example_classes, [3, 5])
        right_right = Node(example, example_classes, [4])
        root.children = [right_child, left_child]
        right_child.children = [right_left, right_right]
        kmdt = KMeansDecisionTree()
        kmdt.fit(example, example_classes)
        assert kmdt.root == root
        assert kmdt.root.left_child == left_child or kmdt.root.right_child == left_child
        assert kmdt.root.left_child == right_child or kmdt.root.right_child == right_child
        if kmdt.root.left_child.left_child is not None and kmdt.root.left_child.right_child is not None:
            assert (
                       kmdt.root.left_child.left_child == right_left and kmdt.root.left_child.right_child == right_right) or (
                       kmdt.root.left_child.right_child == right_left and kmdt.root.left_child.left_child == right_right)
            assert kmdt.root.right_child.left_child is None and kmdt.root.right_child.right_child is None
        elif kmdt.root.right_child.left_child is not None and kmdt.root.right_child.right_child is not None:
            assert (
                       kmdt.root.right_child.left_child == right_left and kmdt.root.right_child.right_child ==
                       right_right) or (
                       kmdt.root.right_child.right_child == right_left and kmdt.root.right_child.left_child == right_right)
            assert kmdt.root.left_child.left_child is None and kmdt.root.left_child.right_child is None

    def test_nb_leaves(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        root = Node(example, example_classes, [0, 1, 2, 3, 4, 5])
        left_child = Node(example, example_classes, [0, 1, 2])
        right_child = Node(example, example_classes, [3, 4, 5])
        right_left = Node(example, example_classes, [3, 5])
        right_right = Node(example, example_classes, [4])
        root.children = [right_child, left_child]
        right_child.children = [right_left, right_right]
        kmdt = KMeansDecisionTree()
        kmdt.root = root
        assert kmdt.n_leaves == 3

    def test_classes_by_level(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        kmdt = KMeansDecisionTree()
        kmdt.fit(example, example_classes)
        obtained_classes = kmdt.classes_by_level(example)
        assert (obtained_classes[0][0] == numpy.array([0, 1, 2, 3, 4, 5])).all()
        assert ((obtained_classes[1][0] == numpy.array([0, 1, 2])).all() and (
            obtained_classes[1][1] == numpy.array([3, 4, 5])).all()) or (
                   (obtained_classes[1][1] == numpy.array([0, 1, 2])).all() and (
                       obtained_classes[1][0] == numpy.array([3, 4, 5])).all())
        assert ((obtained_classes[2][0] == numpy.array([4])).all() and (
            obtained_classes[2][1] == numpy.array([3, 5])).all()) or (
                   (obtained_classes[2][1] == numpy.array([4])).all() and (
                       obtained_classes[2][0] == numpy.array([3, 5])).all())

    def test_predict(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 1, 1])
        kmdt = KMeansDecisionTree()
        kmdt.fit(example, example_classes)
        assert (kmdt.predict(numpy.array([[-2, -2]])) == numpy.array([0])).all()

    def test_node_count(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        kmdt = KMeansDecisionTree()
        kmdt.fit(example, example_classes)
        assert (kmdt.node_count == 5)

    def test_depth(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        kmdt = KMeansDecisionTree()
        kmdt.fit(example, example_classes)
        assert (kmdt.depth == 3)

    ### TESTS FORET
    '''
    Ce qui n'est pas teste : node_count/predict/repr
    '''

    def test_init_forest(self):
        kmrf = KMeansForest()
        assert kmrf.n_estimators == 10
        assert (kmrf.estimators == numpy.array([])).all()
        assert kmrf.max_features == 0.75
        assert kmrf.propagation == 'used'

    def test_reset_forest(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        kmrf = KMeansForest()
        kmrf.fit(example, example_classes)
        kmrf.reset()
        assert (kmrf.estimators == numpy.array([])).all()

    def test_forest_predict(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [3, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 0, 1])
        kmrf = KMeansForest()
        with self.assertRaises(ValueError):
            kmrf.predict(example)
        kmrf.fit(example, example_classes)
        assert (kmrf.predict(example) == numpy.array([0, 0, 0, 1, 0, 1])).all()

    def test_forest_fit(self):
        example = numpy.array([[-1, -1], [-2, -1], [-1, -2], [1, 1], [2, 1], [1, 2]])
        example_classes = numpy.array([0, 0, 0, 1, 1, 1])
        kmrf = KMeansForest(n_estimators=1)
        kmrf.fit(example, example_classes)
        tree = kmrf.estimators[0]
        assert tree.root.out is None
        assert tree.root.left_child.out == 0 or tree.root.left_child.out == 1
        assert tree.root.right_child.out == 1 or tree.root.right_child.out == 0
        assert tree.root.left_child.out != tree.root.right_child.out
        assert tree.root.left_child.children == [None, None]
        assert tree.root.right_child.children == [None, None]

